@@ -85,6 +85,7 @@ typedef struct _server_t
     struct sockaddr* host;
     socklen_t hostlen;
 
+    bool got_any_data;
     timecb_t reconnect_timecb;
 
     conn_state_t state;
@@ -576,13 +577,20 @@ static char* find_config(void)
     {
         if (home != NULL && *home != '\0')
         {
-            asprintf(&tmp2, "%s/.config", home);
+            if (asprintf(&tmp2, "%s/.config", home) == -1)
+            {
+                return NULL;
+            }
             dir = tmp2;
         }
     }
     if (dir != NULL && *dir != '\0')
     {
-        asprintf(&tmp, "%s/upnpproxy.conf", dir);
+        if (asprintf(&tmp, "%s/upnpproxy.conf", dir) == -1)
+        {
+            free(tmp2);
+            return NULL;
+        }
         if (access(tmp, R_OK) == 0)
         {
             free(tmp2);
@@ -604,7 +612,11 @@ static char* find_config(void)
         char* token = strtok(dirtmp, ":");
         while (token != NULL)
         {
-            asprintf(&tmp, "%s/upnpproxy.conf", token);
+            if (asprintf(&tmp, "%s/upnpproxy.conf", token) == -1)
+            {
+                free(dirtmp);
+                return NULL;
+            }
             if (access(tmp, R_OK) == 0)
             {
                 free(dirtmp);
@@ -618,7 +630,10 @@ static char* find_config(void)
     }
     if (home != NULL && *home != '\0')
     {
-        asprintf(&tmp, "%s/.upnpproxy.conf", dir);
+        if (asprintf(&tmp, "%s/.upnpproxy.conf", dir) == -1)
+        {
+            return NULL;
+        }
         if (access(tmp, R_OK) == 0)
         {
             return tmp;
@@ -1058,6 +1073,7 @@ static void daemon_lost_server(daemon_t daemon, server_t* srv, bool wait)
         daemon_clear_remotes(daemon, srv);
     }
     srv->state = CONN_DEAD;
+    srv->got_any_data = false;
     if (srv->reconnect_timecb == NULL)
     {
         srv->reconnect_timecb =
@@ -1132,7 +1148,7 @@ static bool parse_location(const char* location, char** proto,
         ++pos;
         errno = 0;
         x = strtoul(pos, &end, 10);
-        if (errno || !end || *end || x < 1 || x >= 65535)
+        if (errno || !end || (*end != '/' && *end != '\0') || x < 1 || x >= 65535)
         {
             return false;
         }
@@ -1177,11 +1193,19 @@ static char* build_location(const char* proto, const struct sockaddr* host,
     *pos = '\0';
     if (addr_is_ipv6(host, hostlen))
     {
-        asprintf(&ret, "%s://[%s]:%u/%s", proto, tmp, port, path);
+        if (asprintf(&ret, "%s://[%s]:%u/%s", proto, tmp, port, path) == -1)
+        {
+            free(tmp);
+            return NULL;
+        }
     }
     else
     {
-        asprintf(&ret, "%s://%s:%u/%s", proto, tmp, port, path);
+        if (asprintf(&ret, "%s://%s:%u/%s", proto, tmp, port, path) == -1)
+        {
+            free(tmp);
+            return NULL;
+        }
     }
     free(tmp);
     return ret;
@@ -1343,7 +1367,7 @@ static void daemon_lost_tunnel(tunnel_t* tunnel)
         if (tunnel->daemon_conn.state >= CONN_DEAD)
         {
             pkg_t pkg;
-            pkg_close_tunnel(&pkg, tunnel->id);
+            pkg_close_tunnel(&pkg, tunnel->id, true);
             daemon_server_write_pkg(tunnel->source.remote.service->source, &pkg, true);
         }
     }
@@ -1353,19 +1377,11 @@ static void daemon_lost_tunnel(tunnel_t* tunnel)
         if (tunnel->daemon_conn.state >= CONN_DEAD)
         {
             pkg_t pkg;
-            pkg_close_tunnel(&pkg, tunnel->id);
+            pkg_close_tunnel(&pkg, tunnel->id, false);
             daemon_server_write_pkg(tunnel->source.local.server, &pkg, true);
         }
     }
 
-    if (tunnel->stasis)
-    {
-        daemon_release_tunnel_port(daemon, tunnel);
-    }
-
-    close_conn(daemon, &tunnel->local_conn);
-    close_conn(daemon, &tunnel->daemon_conn);
-
     if (tunnel->remote)
     {
         map_remove(tunnel->source.remote.service->source->remote_tunnels, tunnel);
@@ -1423,6 +1439,10 @@ static void tunnel_free(tunnel_t* tunnel)
     {
         daemon = tunnel->source.local.server->daemon;
     }
+    if (tunnel->stasis)
+    {
+        daemon_release_tunnel_port(daemon, tunnel);
+    }
     free_conn(daemon, &tunnel->local_conn);
     free_conn(daemon, &tunnel->daemon_conn);
     http_proxy_free(tunnel->proxy);
@@ -1505,7 +1525,7 @@ static bool flush_conn(daemon_t daemon, tunnel_t* tunnel,
                            "%s tunnel %s connection closed before sending %lu bytes of queued data",
                            tunnel->remote ? "Remote" : "Local",
                            in_conn == &tunnel->local_conn ? "local" : "daemon",
-                           buf_ravail(in_conn->buf));
+                           (unsigned long)buf_ravail(in_conn->buf));
             }
             if (proxy != NULL)
             {
@@ -1568,7 +1588,7 @@ static bool flush_conn(daemon_t daemon, tunnel_t* tunnel,
                        "%s tunnel %s connection closed when sending %lu bytes of queued data",
                        tunnel->remote ? "Remote" : "Local",
                        in_conn == &tunnel->local_conn ? "local" : "daemon",
-                       avail);
+                       (unsigned long)avail);
             close_conn(daemon, in_conn);
             return true;
         }
@@ -1937,7 +1957,7 @@ static void daemon_create_tunnel(daemon_t daemon, server_t* server,
     tunnel.local_conn.sock = socket_tcp_connect2(
                                  tunnel.source.local.service->host,
                                  tunnel.source.local.service->hostlen,
-                                 false);
+                                 false, daemon->bind_services);
     if (tunnel.local_conn.sock < 0)
     {
         pkg_t pkg;
@@ -1974,7 +1994,8 @@ static void daemon_create_tunnel(daemon_t daemon, server_t* server,
         addr_setport(host, server->hostlen, create_tunnel->port);
         tunnelptr->daemon_conn.state = CONN_CONNECTING;
         tunnelptr->daemon_conn.sock = socket_tcp_connect2(host, server->hostlen,
-                                                          false);
+                                                          false,
+                                                          daemon->bind_server);
         if (tunnelptr->daemon_conn.sock < 0)
         {
             char* tmp;
@@ -2025,9 +2046,17 @@ static void daemon_close_tunnel(daemon_t daemon, server_t* server,
 {
     tunnel_t key;
     key.id = close_tunnel->tunnel_id;
-    key.remote = false;
-    key.source.local.server = server;
-    map_remove(server->local_tunnels, &key);
+    if (close_tunnel->local)
+    {
+        key.remote = false;
+        key.source.local.server = server;
+        map_remove(server->local_tunnels, &key);
+    }
+    else
+    {
+        key.remote = true;
+        map_remove(server->remote_tunnels, &key);
+    }
 }
 
 static void daemon_setup_tunnel(daemon_t daemon, server_t* server,
@@ -2074,7 +2103,8 @@ static void daemon_setup_tunnel(daemon_t daemon, server_t* server,
 
         tunnel->daemon_conn.state = CONN_CONNECTING;
         tunnel->daemon_conn.sock = socket_tcp_connect2(host, server->hostlen,
-                                                       false);
+                                                       false,
+                                                       daemon->bind_server);
         if (tunnel->daemon_conn.sock < 0)
         {
             char* tmp;
@@ -2120,9 +2150,18 @@ static void daemon_server_incoming_cb(void* userdata, socket_t sock)
             }
 
             asprinthost(&tmp, server->host, server->hostlen);
-            log_printf(daemon->log, LVL_WARN,
-                       "Unable to connect to server %s: %s",
-                       tmp, socket_strerror(sock));
+            if (got == 0)
+            {
+                log_printf(daemon->log, LVL_WARN,
+                           "Unable to connect to server %s: Connection closed",
+                           tmp);
+            }
+            else
+            {
+                log_printf(daemon->log, LVL_WARN,
+                           "Unable to connect to server %s: %s",
+                           tmp, socket_strerror(sock));
+            }
             free(tmp);
             daemon_lost_server(daemon, server, true);
             return;
@@ -2166,7 +2205,7 @@ static void daemon_server_incoming_cb(void* userdata, socket_t sock)
                                tmp,
                                socket_strerror(sock));
                     free(tmp);
-                    daemon_lost_server(daemon, server, false);
+                    daemon_lost_server(daemon, server, !server->got_any_data);
                     return;
                 }
             }
@@ -2178,11 +2217,12 @@ static void daemon_server_incoming_cb(void* userdata, socket_t sock)
                            "Lost connection with server %s: Connection closed",
                            tmp);
                 free(tmp);
-                daemon_lost_server(daemon, server, false);
+                daemon_lost_server(daemon, server, !server->got_any_data);
                 return;
             }
             else
             {
+                server->got_any_data = true;
                 buf_wmove(server->in, got);
             }
         }
@@ -2223,6 +2263,24 @@ static void daemon_server_incoming_cb(void* userdata, socket_t sock)
 
 static int _daemon_server_flush_output(server_t* server);
 
+static void daemon_server_connected(server_t* server)
+{
+    size_t i;
+    daemon_t daemon = server->daemon;
+    assert(server->state == CONN_CONNECTED);
+
+    for (i = map_begin(daemon->locals); i != map_end(daemon->locals);
+         i = map_next(daemon->locals, i))
+    {
+        pkg_t pkg;
+        localservice_t* local = map_getat(daemon->locals, i);
+        pkg_new_service(&pkg, local->id, local->usn, local->location,
+                        local->service, local->server, local->opt,
+                        local->nls);
+        daemon_server_write_pkg(server, &pkg, false);
+    }
+}
+
 static void daemon_server_writable_cb(void* userdata, socket_t sock)
 {
     server_t* server = userdata;
@@ -2235,23 +2293,9 @@ static void daemon_server_writable_cb(void* userdata, socket_t sock)
     case CONN_DEAD:
         return;
     case CONN_CONNECTING:
-    {
-        size_t i;
-        daemon_t daemon = server->daemon;
         server->state = CONN_CONNECTED;
-
-        for (i = map_begin(daemon->locals); i != map_end(daemon->locals);
-             i = map_next(daemon->locals, i))
-        {
-            pkg_t pkg;
-            localservice_t* local = map_getat(daemon->locals, i);
-            pkg_new_service(&pkg, local->id, local->usn, local->location,
-                            local->service, local->server, local->opt,
-                            local->nls);
-            daemon_server_write_pkg(server, &pkg, false);
-        }
+        daemon_server_connected(server);
         break;
-    }
     case CONN_CONNECTED:
         break;
     }
@@ -2319,6 +2363,7 @@ static void daemon_server_accept_cb(void* userdata, socket_t sock)
                 }
                 daemon->server[i].state = CONN_CONNECTED;
                 daemon->server[i].sock = s;
+                daemon_server_connected(daemon->server + i);
                 socket_setblocking(s, false);
                 selector_add(daemon->selector, daemon->server[i].sock,
                              daemon->server + i,
@@ -2326,6 +2371,7 @@ static void daemon_server_accept_cb(void* userdata, socket_t sock)
                              daemon_server_writable_cb);
                 selector_chkwrite(daemon->selector, daemon->server[i].sock,
                                   false);
+                daemon_server_flush_output(daemon->server + i);
                 break;
             case CONN_CONNECTING:
                 selector_remove(daemon->selector, daemon->server[i].sock);
@@ -2389,7 +2435,8 @@ static bool daemon_setup_remote_server(daemon_t daemon, server_t* srv)
         srv->out = buf_new(SERVER_BUFFER_OUT);
     }
     srv->state = CONN_CONNECTING;
-    srv->sock = socket_tcp_connect2(srv->host, srv->hostlen, false);
+    srv->sock = socket_tcp_connect2(srv->host, srv->hostlen, false,
+                                    daemon->bind_server);
     if (srv->sock < 0)
     {
         char* tmp;
@@ -2493,7 +2540,6 @@ bool load_config(daemon_t daemon)
         return false;
     }
 
-    cfg_close(cfg);
 
     if (safestrcmp(bind_multicast, daemon->bind_multicast) != 0)
     {
@@ -2559,6 +2605,10 @@ bool load_config(daemon_t daemon)
                 daemon->tunnel_port = realloc(daemon->tunnel_port, nc * sizeof(tunnel_port_t));
                 memset(daemon->tunnel_port + i, 0,
                        (nc - i) * sizeof(tunnel_port_t));
+                for (j = i; j < nc; ++j)
+                {
+                    daemon->tunnel_port[j].sock = -1;
+                }
                 daemon->tunnel_port_count = nc;
             }
         }
@@ -2654,6 +2704,8 @@ bool load_config(daemon_t daemon)
         server_cnt = 0;
     }
 
+    cfg_close(cfg);
+
     return true;
 }
 
@@ -2744,13 +2796,19 @@ static char* daemon_generate_uid(daemon_t daemon)
             }
             if (home != NULL && *home != '\0')
             {
-                asprintf(&tmp2, "%s/.cache", home);
+                if (asprintf(&tmp2, "%s/.cache", home) == -1)
+                {
+                    return NULL;
+                }
                 dir = tmp2;
             }
         }
         if (dir != NULL && *dir != '\0')
         {
-            asprintf(&tmp, "%s/upnpproxy.cache", dir);
+            if (asprintf(&tmp, "%s/upnpproxy.cache", dir) == -1)
+            {
+                return NULL;
+            }
             free(tmp2);
         }
         if (tmp != NULL)
@@ -3111,6 +3169,10 @@ static int _daemon_server_flush_output(server_t* server)
     size_t avail;
     const char* ptr;
     ssize_t got;
+    if (server->state != CONN_CONNECTED)
+    {
+        return 1;
+    }
     for (;;)
     {
         ptr = buf_rptr(server->out, &avail);
@@ -3122,7 +3184,7 @@ static int _daemon_server_flush_output(server_t* server)
         if (got <= 0)
         {
             char* tmp;
-            if (socket_blockingerror(server->sock))
+            if (socket_blockingerror(server->sock) || got == 0)
             {
                 return 1;
             }
